/*
 *	adsbox - ADS-B decoder software
 *	(c) atty 2011-14 romulmsg@mail.ru
 *
 *	Nikolay Zlatev code rework
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include "tcp_beast.h"
#include "misc.h"
#include "adsb.h"
#include "db.h"

extern sqlite3 *db;

int db_insert_source_tcp_beast(unsigned int *id, char *ip) {

	if (db_exec_sql("INSERT INTO SOURCE_TCP_BEAST (ID, IP) VALUES (%i, \"%s\")",
			*id, ip)) return 1;

	return 0;
}

int db_update_source_tcp_beast_port(unsigned int *id, char *port) {

	if (db_exec_sql("UPDATE SOURCE_TCP_BEAST SET PORT=\"%s\" WHERE ID=%i", port, *id))
	    return 1;

	if (!sqlite3_changes(db)) {
		DEBUG("Can not set TCP port for source id %u\n", *id);
		return 1;
	}

	return 0;
}

int run_source_tcp_beast() {
	char ** results;
	int rows, cols, i;
	char *zErrMsg = 0;
	char sql[] = "SELECT ID, IP, PORT, LAT, LON FROM SOURCE_TCP_BEAST ORDER BY ID";
	pthread_t child;

	if (sqlite3_get_table(db, sql, &results, &rows, &cols, &zErrMsg) != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		sqlite3_free_table(results);
		return 1;
	}

	for (i = 1; i <= rows; i++) {
		tcp_beast_info *ti = malloc(sizeof(tcp_beast_info));
		ti->id = atoi(results[i * cols]);
		strncpy(ti->addr, results[i * cols + 1], sizeof(ti->addr));
		strncpy(ti->port,results[i * cols + 2], sizeof(ti->port));
		ti->lat = atof(results[i * cols + 3]);
		ti->lon = atof(results[i * cols + 4]);
		pthread_create(&child, NULL, &tcp_beast_client, ti); /* start tcp client thread */
		pthread_detach(child);
	}

	sqlite3_free_table(results);
	return 0;
}

void *tcp_beast_client(void *arg) {
	tcp_beast_info *ti = (tcp_beast_info*) arg;
	unsigned char beast_data[23];
	char name[64];
	unsigned char idx = 0, c = 0, c_prev = 0, msg_len = 0;
	source_stat status;
	adsb_data *data = malloc(sizeof(adsb_data));
	clockid_t cid;
	int fd, len = 0;
	struct addrinfo hints, *res;

	pthread_getcpuclockid(pthread_self(), &cid);

	data->source_id = ti->id;
	data->type = TYPE_TCP_BEAST; // type of source
	data->lat = ti->lat;
	data->lon = ti->lon;
	data->data_len = 0;
	bzero(&status, sizeof(source_stat));
	data->status = &status;

	db_stat_addr(ti->id, &status);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	if (getaddrinfo(ti->addr, ti->port, &hints, &res) != 0) {
		panic("getaddrinfo");
	}

	fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (fd < 0)
		panic("socket");

	sprintf(name, "Beast TCP  %s:%s", ti->addr, ti->port);
	db_insert_thread(cid, name);

	DEBUG("Connecting to beast binary data server %s:%s...\n", ti->addr,
			ti->port);

	if (connect(fd, res->ai_addr, res->ai_addrlen) < 0) {
		DEBUG("Could not connect to %s:%s\n", ti->addr, ti->port);
		goto exit;
	}

	unsigned int state = STATE_UNKNOWN, freq, nl;
	DEBUG("Connected to %s:%s\n", ti->addr, ti->port);
	while ((len = read(fd, &c, 1)) > 0) {
		if (((c != ESC) && (c_prev == ESC)) || (idx == sizeof(beast_data))) { // esc byte received - start of frame

			if (state == STATE_ADSB) {
				memcpy(data->data, (char *) &beast_data + idx - msg_len - 1,
					msg_len);
				data->timestamp = beast_data[6] + (beast_data[5] << 8) +
					(beast_data[4] << 16) + (beast_data[3] << 24);
				if (data->msg_type == MSG_TYPE_SHORT)
					sprintf(data->avr_data, "*%02X%02X%02X%02X%02X%02X%02X;",
						beast_data[8], beast_data[9],beast_data[10],beast_data[11],
						beast_data[12],beast_data[13],beast_data[14]);
				if (data->msg_type == MSG_TYPE_LONG)
					sprintf(data->avr_data,
						"*%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X;",
						beast_data[8], beast_data[9],beast_data[10],beast_data[11],
						beast_data[12],beast_data[13],beast_data[14],
						beast_data[15], beast_data[16],beast_data[17],beast_data[18],
						beast_data[19],beast_data[20],beast_data[21]);
				enqueue_data(data);
			}
/*
	ADSBox beast protocol extension, new packet type
	<esc>, 0x35, 4 byte MLAT counter frequency, 1 byte noise level
*/
			if (state == STATE_TELEMETRY) {
				freq = beast_data[4] + (beast_data[3] << 8) +
					(beast_data[2] << 16) + (beast_data[1] << 24);
				nl = beast_data[5];
				db_update_hz(data->source_id, freq);
				db_update_nl(data->source_id, nl);
			}

			idx = 0;
			data->data_len = 0;
		}

/*
	Beast binary protocol is defined at
	http://wiki.modesbeast.com/Mode-S_Beast:Data_Output_Formats
*/

		if ((c_prev == ESC) && (c == '2') && (idx==0)) {
			data->data_len = 16;
			data->msg_type = MSG_TYPE_SHORT;
			msg_len = 7;
			state = STATE_ADSB;
		}
		if ((c_prev == ESC) && (c == '3') && (idx==0)) {
			data->data_len = 23;
			data->msg_type = MSG_TYPE_LONG;
			msg_len = 14;
			state = STATE_ADSB;
		}
		if ((c_prev == ESC) && (c == '5') && (idx==0)) {
			data->data_len = 7;
			data->msg_type = MSG_TYPE_TELEMETRY;
			state = STATE_TELEMETRY;
		}
		if ((c_prev == ESC) && (c == ESC)) { // ESC stuffing
			beast_data[idx] = ESC;
			idx--;
			c = 0;
		} else
			beast_data[idx] = c;

		idx++;
		c_prev = c;
	}

	exit:
	db_delete_thread(cid);
	db_del_source(ti->id);
	free(ti);
	free(data);
	if (res)
		freeaddrinfo(res);
	DEBUG("\ntcp_beast_client exit");
	pthread_exit(0);
}
